/* 
  skip.c    */

#include <stdio.h>
#include <string.h>
#include "skip.h"

/* the header of a skip list file identifies the file and
   describes the sizes of the components of the file */

struct skip_header {
    char magic[4];		/* "SKIP" */
    char version[4];		/* "001\0" */
    int table_size;		/* number of entries */
    int key_size;		/* size of the string table */
    int data_size;		/* size of the user data */
};

#define SKIP_MAGIC "SKIP"
#define SKIP_VERSION "001"	/* we store all four bytes */

#define MAX_LEVELS 16		/* allows LEVEL_STEP^MAX_LEVELS entries */
#define LEVEL_STEP 4		/* 4^16 = 10^9 */

/* this macro allocates a new node structure.  note that
   a level 0 node needs one forwarding slot */
#define new_node(l) ((node)malloc(sizeof(_node_el)+l*sizeof(_node_el)))

/* copier for normal strings.  the default skip list is a mapping from
   strings to ints */
static int string_copy(char *dst, char *src)
{
    strcpy(dst,src);
    return strlen(src)+1;
}

static int string_size(char *s)
{
    return strlen(s)+1;		/* no alignment needed for strings */
}

/* create a skip list which maps strings to ints */
typedef int (*intfun)();

skip_list create_list_with_ops(intfun k_cmp,
			       intfun k_copy, intfun k_size, intfun k_free,
			       intfun v_copy, intfun v_size, intfun v_free)
{
    skip_list l;
    int i;

    l = (skip_list) malloc(sizeof(*l));
    l->level = 0;
    l->compare = k_cmp;
    l->key_copy = k_copy;
    l->key_size = k_size;
    l->key_free = k_free;
    l->copy = v_copy;
    l->size = v_size;
    l->free = v_free;
    l->header = new_node(MAX_LEVELS);
    if (!l->header) {
	fprintf(stderr, "could not allocate %d bytes\n",
		sizeof(struct s_node)+MAX_LEVELS*sizeof(struct s_node));
	exit(1);
    }
    for (i=0;i<MAX_LEVELS;i++) {
	l->header->forward[i] = NULL;
    }
    return l;
}

skip_list create_skip_list()
{
    return create_list_with_ops(strcmp, string_copy, string_size, NULL,
				NULL, /* implies no space is needed */
				NULL, /* implies bitwise 4 byte copy of v */
				NULL);
}

/* deallocates a skip list as thoroughly as we know how */
void free_skip_list( skip_list l )
{
    node p, q;
    p = l->header;
    while (p) {
	q = p->forward[0];
	if (l->key_free) l->key_free(p->key);
	if (l->free) l->free(p->v);
	free(p);
	p = q;
    }
    free(l);
}

/* insert a key/value into a skip list.  if replace_dups is set,
   then any previous entry with the same key will be overwritten.
   otherwise, new entries precede any previous entries */
int insert(skip_list l, void *key, value v, int replace_dups)
{
    int i,k;
    node p,q;
    node update[MAX_LEVELS];

    /* first, we skip across at the various levels until we find
       the entry which is just to the left of the desired place
       on each level.  we start at the top to take big jumps. */
    q = NULL;
    p = l->header;
    for (i=l->level-1;p && i>=0;i--) {
	q = p->forward[i];
	while (q && l->compare(q->key,key) < 0) {
	    p = q;
	    q = p->forward[i];
	}
	update[i] = p;
    }

    /* so now update contains pointers to all the nodes just left of
       us on all levels and q may just point to the sought after node */    
    if (replace_dups && q && strcmp(q->key, key) == 0) {
	if (l->free) l->free(q->v);
	q->v = v;
	return 0;
    }

    k=0;			/* start low, then increment a bit, maybe */
    while (k < l->level && (random()%LEVEL_STEP) == 0) k++;

    if (k >= l->level) {
	l->level++;		/* may need to increase level of entire list */
	update[k] = l->header;
    }
    q = new_node(k);
    if (!q) {
	fprintf(stderr, "could not allocate %d bytes\n",
		sizeof(struct s_node)+k*sizeof(struct s_node));
	exit(1);
    }
    if (l->key_copy && l->key_size) {
	q->key = (void *) emalloc(l->key_size(key));
	l->key_copy(q->key, key);
    }
    else q->key = key;

    if (l->copy && l->size) {
	q->v = emalloc(l->size(v));
	l->copy(q->v, v);
    }
    else q->v = v;

    for (i=0;i<=k;i++) {
	p = update[i];
	q->forward[i] = p->forward[i];
	p->forward[i] = q;
    }
    return 1;
}

void *first_key(skip_list l)
{
    return l.header->forward[0]->key;
}

/* deletes an item from a list... returns 1 if the item was in the list,
   0 if nothing was found */
int delete(skip_list l, void *key)
{
    int i,k;
    node p,q;
    node update[MAX_LEVELS];

    p = l->header;
    q = NULL;
    for (i=l->level-1;p && i>=0;i--) {
	q = p->forward[i];
	while (q && l->compare(q->key,key) < 0) {
	    p = q;
	    q = p->forward[i];
	}
	update[i] = p;
    }

    if (l->compare(q->key, key) == 0) {
	for (i=0;i<l->level && update[i]->forward[i] == q;i++) {
	    update[i]->forward[i] = q->forward[i];
	}
	if (l->key_free) l->key_free(q->key);
	if (l->free) l->free(q->v);
	free(q);
	i = l->level;
	while (l->header->forward[i] == NULL) i--;
	l->level = i;
	return 1;
    }
    else return 0;
}

static node finger = NULL;	/* saved for find_next */
static skip_list last_l;

int find(skip_list l, void *key, value *v)
{
    int i,k;
    node p,q;

    last_l = l;
    p = l->header;
    for (i=l->level-1;i>=0;i--) {
	if (p) q = p->forward[i];
	else q = NULL;
	while (q && l->compare(q->key,key) < 0) {
	    p = q;
	    q = p->forward[i];
	}
    }

    if (l->compare(q->key, key) != 0) return 0;
    else {
	if (l->copy && l->size) {
	    *v = emalloc(l->size(q->v));
	    l->copy(*v, q->v);
	}
	else *v = q->v;
	finger = q;
	return 1;
    }
}

int find_next(skip_list l, void *key, value *v)
{
    if (l != last_l) {
	return find(l, key, v);
    }
    else {
	finger = finger->forward[0];
	if (finger && l->compare(finger->key, key) == 0) {
	    if (l->copy && l->size) {
		*v = emalloc(l->size(finger->v));
		l->copy(*v, finger->v);
	    }
	    else *v = finger->v;
	    return 1;
	}
	else {
	    return 0;
	}
    }
}

void modify(skip_list l, void *key, value v)
{
    if (l == last_l && finger && l->compare(finger->key, key) == 0) {
	if (l->free) l->free(finger->v);
	if (l->copy && l->size) {
	    finger->v = l->size(v);
	    l->copy(finger->v, v);
	}
	else {
	    finger->v = v;
	}
    }
    else {
	value w;
	if (find(l, key, &w)) {
	    modify(l, key, v);
	}
	else {
	    fprintf(stderr, "cannot modify element that isn't in list\n");
	    exit(1);
	}
    }
}

static int compare_address(void *a, void *b)
{
    unsigned int aa, bb;
    aa = (unsigned) a;
    bb = (unsigned) b;
    return bb-aa;
}

void write_skip_list(FILE *f, skip_list l)
{
    int i;
    int table_count;
    int key_size, data_size;
    int *key_offsets;
    skip_list data_tab;
    char *key_space, *pk, *data_space, *pd;
    node p;
    struct skip_header header;

    /* data address -> data offsets */
    data_tab = create_list_with_ops(compare_address,
				    NULL, NULL, NULL, 
				    NULL, NULL, NULL);

    key_size = 0;
    data_size = 0;
    table_count = 0;
    for (p = l->header->forward[0];p;p = p->forward[0]) {
	if (l->key_copy && l->key_size) {
	    key_size += l->key_size(p->key);
	}
	if (l->copy && l->size && insert(data_tab, (void *) p->v, 0, 1)) {
	    data_size += l->size(p->v);
	}
	table_count++;
    }
    
    key_offsets = (int *) ecalloc(table_count, sizeof(key_offsets[0]));
    if (key_size) {
	key_space = (char *) emalloc(key_size);
	p = l->header->forward[0];
	pk = key_space;
	for (i=0;i<table_count;i++) {
	    key_offsets[i] = pk-key_space;
	    pk += l->key_copy(pk, p->key);
	    p = p->forward[0];
	}
    }
    else {
	key_space = NULL;
	p = l->header->forward[0];
	for (i=0;i<table_count;i++) {
	    key_offsets[i] = (int) p->key;
	    p = p->forward[0];
	}
    }

    if (data_size) {
	data_space = (char *) emalloc(data_size);
	pd = data_space;
	for (p = l->header->forward[0];p;p = p->forward[0]) {
	    insert(data_tab, (void *) p->v, pd-data_space, 1);
	    pd += l->copy(pk, p->v);
	}
    }

    strncpy(header.magic, SKIP_MAGIC, 4);
    strncpy(header.version, SKIP_VERSION, 4);
    header.key_size = key_size;
    header.table_size = table_count;
    header.data_size = data_size;
    fwrite(&header, sizeof(header), 1, f);

    for (i=0;i<table_count;i++) {
	fwrite(&key_offsets[i], sizeof(key_offsets[i]), 1, f);
    }
    if (key_size) {
	fwrite(key_space, key_size, 1, f);
    }


    if (data_size) {
	int offset;
	for (p=l->header->forward[0];p;p=p->forward[0]) {
	    find(data_tab, (void *) p->v, &offset);
	    fwrite(&offset, sizeof(offset), 1, f);
	}
	fwrite(data_space, data_size, 1, f);
    }
    else {
	for (p = l->header->forward[0];p;p = p->forward[0]) {
	    fwrite(&p->v, sizeof(p->v), 1, f);
	}
    }
}

skip_list read_skip_list(FILE *f,
			 int (*key_compare)(),
			 int (*key_copy)(), int (*key_size)(),
			 int (*data_copy)(), int (*data_size)())
{
    int i;
    skip_list r;
    int *key_offsets, *data_offsets;
    char *key_space, *data_space;
    struct skip_header header;

    fread(&header, sizeof(header), 1, f);

    key_offsets = (int *) ecalloc(header.key_size+1, sizeof(key_offsets[0]));
    fread(key_offsets, header.table_size, sizeof(key_offsets[0]), f);
    key_offsets[header.table_size] = header.key_size;
    key_space = (char *) emalloc(header.key_size);
    fread(key_space, header.key_size, 1, f);

    data_offsets = (int *) ecalloc(header.table_size+1,
				   sizeof(data_offsets[0]));
    fread(data_offsets, header.table_size, sizeof(data_offsets[0]), f);

#if 0
    for (i=0;i<header.table_size;i++) {
	fread(data_offsets+i, 1, sizeof(data_offsets[0]), f);
	printf("%d\n", i);
    }
#endif
    data_offsets[header.table_size] = header.data_size;

    data_space = (char *) emalloc(header.data_size);
    if (header.data_size) fread(data_space, header.data_size, 1, f);

    r = create_list_with_ops(key_compare,
			     key_copy, key_size, NULL,
			     data_copy, data_size, NULL);
    for (i=0;i<header.table_size;i++) {
	insert(r, key_space+key_offsets[i],
	       (value) data_space+data_offsets[i], 0);
    }
    free(key_offsets);
    free(data_offsets);
    return r;
}

#define LIMIT 10000

main()
{
    FILE *f;
    int i, k;
    value v;
    char keys[LIMIT][20];
    skip_list l, m;
    
    printf("creating file ... ");
    fflush(stdout);
    l = create_skip_list();
    for (k=0;k<LIMIT;k++) {
	sprintf(keys[k], "%d", random());
	if (!insert(l, keys[k], k, 0)) {
	    printf("duplicate key %s\n", keys[k]);
	}
    }
    printf("created.\n");

    printf("writing file ... "); fflush(stdout);
    f = fopen("xyz", "w");
    write_skip_list(f, l);
    fclose(f);
    printf("written.\n");

    printf("reading file ... ");
    fflush(stdout);
    f = fopen("xyz", "r");
    m = read_skip_list(f, strcmp, string_copy, string_size, NULL, NULL);
    printf("read.\n");

    printf("checking file ... ");
    fflush(stdout);
    for(k=0;k<LIMIT;k++) {
	if (!find(m,keys[k],&v)) {
	    printf("error in search #%d,#%d\n",i,k);
	}
	else {
	    if (v != k) printf("\nsearch of restored file copy returned wrong value %d vs %d\n",
			       v, k);
	}
    }
    printf("checked.\n");


    for(i=0;i<4;i++) {
        for(k=0;k<LIMIT;k++) {
            if (!find(l,keys[k],&v)) printf("error in search #%d,#%d\n",i,k);
            else {
		if (v != k) printf("search returned wrong value %d vs %d\n",
				   v, k);
	    }
        }
        for(k=0;k<LIMIT;k++) {
            if (! delete(l,keys[k])) printf("error in delete\n");
            sprintf(keys[k], "%d", random());
            insert(l,keys[k],k,0);
        }
    }
}
