/***************************************************************
 * breakpts.c -- This file contains all functions for manipu-  *
 *               lating breakpoints for CS273's Sim6811_GUI.   *
 *                                                             *
 *                 written by Rob Sinclair                     *
 *                  rsinclai@cs.nmsu.edu                       *
 ***************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/types.h>
#include <xview/xview.h>
#include <xview/panel.h>
#include <xview/textsw.h>
#include <xview/xv_xrect.h>
#include <gfm.h>
#include "sim6811_ui.h"
#include "defs.h"

/*****************************************************************
 * Breakpoints are controlled via the following rules:           *
 *   CLEAR   ==  0 -> No Breakpoint                              *
 *   PERM    == -1 -> Permanent Breakpoint                       *
 *   TEMP(n) == +n -> Temporary Breakpoint that will evaporate   *
 *                     after being hit n times.                  *
 *****************************************************************/
#define CLEAR    0
#define PERM    -1
#define TEMP(n)  n

extern  short   breaks[65536];

extern  sim6811_sim6811Window_objects	*Sim6811_sim6811Window;
extern  sim6811_winSourceCode_objects	*Sim6811_winSourceCode;
extern  sim6811_winCrossRef_objects	*Sim6811_winCrossRef;
extern  sim6811_winCircuitBoard_objects	*Sim6811_winCircuitBoard;
extern  sim6811_winMemory_objects	*Sim6811_winMemory;
extern  sim6811_winRegisters_objects	*Sim6811_winRegisters;
extern  sim6811_winStack_objects	*Sim6811_winStack;
extern  sim6811_winSymbolTable_objects	*Sim6811_winSymbolTable;
extern  sim6811_winOutput_objects	*Sim6811_winOutput;
extern  sim6811_winEditMemory_objects	*Sim6811_winEditMemory;
extern  sim6811_winWarning_objects	*Sim6811_winWarning;
extern  sim6811_winBreakpoints_objects	*Sim6811_winBreakpoints;

/****************************************************************
 * Called by the callback function for `btnBreakPtsAdd'.
 ****************************************************************/
void newBreakPt()
{
    int  address, type, numTimes;
    char text[80], addr[10];

    /* Get information for the new breakpoint */
    strcpy(addr, (char*)xv_get(Sim6811_winBreakpoints->txtBreakPtsAddress, 
			       PANEL_VALUE, NULL));
    sscanf(addr, "%x", &address);
    type = (int)xv_get(Sim6811_winBreakpoints->setBreakPtsType, 
			  PANEL_VALUE, NULL);

    if (address>=0 && address<65536) {
	/* If a breakpoint is already set at this address, abort. */
	if (breaks[address] != 0) {
	    fprintf(stderr, "\a");
	    output(WIN_OUTPUT, "Invalid Breakpoint:  %s%04x.\n",
		   "A breakpoint already exists at address 0x",
		   address);
	    return;
	}

	/* Build the string to display in the breakpoint window and
	   add the new breakpoint to the simulator's state. */
	if ( type ) { /* If Temporary */
	    numTimes = (int)xv_get(Sim6811_winBreakpoints->txtBreakPtTimes, 
				   PANEL_VALUE, NULL);

	    /* Create the new temporary breakpoint */
	    breaks[address] = TEMP(numTimes);
	    
	    sprintf(text, "0x%04x (Temporary: %d times)", address, numTimes);
	}
	else {
	    /* Create the new permanent breakpoint */
	    breaks[address] = PERM;
	    
	    sprintf(text, "0x%04x (Permanent)", address);
	}

	/* Display the new breakpoint in the list and resort the list */
	xv_set(Sim6811_winBreakpoints->listBreakPts, 
	       PANEL_LIST_INSERT, 0, 
	       PANEL_LIST_STRING, 0, text, 
	       PANEL_LIST_SORT, PANEL_FORWARD, NULL);
	
	/* Activate the "Update" and "Delete" buttons */
	xv_set(Sim6811_winBreakpoints->btnBreakPtsUpdate, 
	       PANEL_INACTIVE, FALSE, NULL); 
	xv_set(Sim6811_winBreakpoints->btnBreakPtsDelete, 
	       PANEL_INACTIVE, FALSE, NULL); 
    }
}

/****************************************************************
 * Called by the callback function for `btnBreakPtsUpdate'.  This
 * allows the user to edit the Address, Type,  and # of occurances
 * of each breakpoint.
 ****************************************************************/
void editBreakPt()
{
    int  selection, oldAddress, newAddress, type, numTimes;
    char text[80], oldAddr[80], newAddr[80];

    /* Find out which item is selected */
    selection = (int)xv_get(Sim6811_winBreakpoints->listBreakPts, 
			    PANEL_LIST_FIRST_SELECTED, NULL);

    if (selection != -1) {
	/* Get address from the selected breakpoint */
	strcpy(oldAddr,(char*)xv_get(Sim6811_winBreakpoints->listBreakPts, 
				   PANEL_LIST_STRING, selection, NULL));
	sscanf(oldAddr, "0x%x", &oldAddress);

	/* Clear the old breakpoint from memory. */
	breaks[oldAddress] = CLEAR;

	/* Get the new breakpoint's address. */
	strcpy(newAddr, 
	       (char*)xv_get(Sim6811_winBreakpoints->txtBreakPtsAddress, 
			     PANEL_VALUE, NULL));
	sscanf(newAddr, "%x", &newAddress);

	/* Get Type and/or Occurances for the new breakpoint */
	type = (int)xv_get(Sim6811_winBreakpoints->setBreakPtsType, 
			   PANEL_VALUE, NULL);
	
	/* Build the string to display in the breakpoint window and
	   change the breakpoint in the simulator's state. */
	if ( type ) { /* If Temporary */
	    numTimes = (int)xv_get(Sim6811_winBreakpoints->txtBreakPtTimes, 
				   PANEL_VALUE, NULL);

	    /* Create the new temporary breakpoint */
	    breaks[newAddress] = TEMP(numTimes);

	    sprintf(text, "0x%04x (Temporary: %d times)", 
		    newAddress, numTimes);
	}
	else {
	    /* Create the new permanent breakpoint */
	    breaks[newAddress] = PERM;
	    
	    sprintf(text, "0x%04x (Permanent)", newAddress);
	}

	/* Display the new breakpoint in the list */
	xv_set(Sim6811_winBreakpoints->listBreakPts, 
	       PANEL_LIST_DELETE, selection, 
	       PANEL_LIST_INSERT, selection, 
	       PANEL_LIST_STRING, selection, text, 
	       PANEL_LIST_SORT, PANEL_FORWARD, NULL);
    }
}

/****************************************************************
 * Called by the callback function for `btnBreakPtsDelete'.
 ****************************************************************/
void removeBreakPt()
{
    int  numRows, selection, address, type, numTimes;
    char text[80], addr[80];

    /* Find out which item is selected and how many items exist */
    selection = (int)xv_get(Sim6811_winBreakpoints->listBreakPts, 
			    PANEL_LIST_FIRST_SELECTED, NULL);
    numRows = (int)xv_get(Sim6811_winBreakpoints->listBreakPts, 
			  PANEL_LIST_NROWS, NULL);

    if (selection != -1) {
	/* Get address from the selected breakpoint */
	strcpy(addr,(char*)xv_get(Sim6811_winBreakpoints->listBreakPts, 
				   PANEL_LIST_STRING, selection, NULL));
	sscanf(addr, "0x%x", &address);

	/* Clear the breakpoint from memory. */
	breaks[address] = CLEAR;

	/* Remove the breakpoint from the list. */
	if (address>=0 && address<65536)
	  xv_set(Sim6811_winBreakpoints->listBreakPts, 
		 PANEL_LIST_DELETE, selection, NULL);

	/* If this breakpoint was the only item in the list, 
	   disable the "Update" and "Delete" buttons. */
	if (numRows-1 == 0) {
	    /* Deactivate the "Update" and "Delete" buttons */
	    xv_set(Sim6811_winBreakpoints->btnBreakPtsUpdate, 
		   PANEL_INACTIVE, TRUE, NULL); 
	    xv_set(Sim6811_winBreakpoints->btnBreakPtsDelete, 
		   PANEL_INACTIVE, TRUE, NULL); 
	}
    }
}


/****************************************************************
 * Called by the callback function for `listBreakPts'.
 ****************************************************************/
void copyBrkPtSelectionInfo(char *string)
{
    int  type=0, numTimes=0;
    char address[10];

    /* Parse the selected string */
    if (sscanf(string, "0x%s (Temporary: %d", address, &numTimes) == 2){
	type = 1; /* Temp breakpoint */

	/* Activate the Occurances field */
	xv_set(Sim6811_winBreakpoints->txtBreakPtTimes, 
	       PANEL_INACTIVE, FALSE, NULL);
    }
    else {
	sscanf(string, "0x%s (Permanent)", address);
	type = 0; /* Perm breakpoint */
	numTimes = 0;

	/* Inactivate the Occurances field */
	xv_set(Sim6811_winBreakpoints->txtBreakPtTimes, 
	       PANEL_INACTIVE, TRUE, NULL);
    }
    
    /* Update the Address, Type and Occurances breakpoint fields */
    xv_set(Sim6811_winBreakpoints->txtBreakPtsAddress, 
	   PANEL_VALUE, address, NULL);
    xv_set(Sim6811_winBreakpoints->setBreakPtsType, 
	   PANEL_VALUE, type, NULL);
    xv_set(Sim6811_winBreakpoints->txtBreakPtTimes, 
	   PANEL_VALUE, numTimes, NULL);
}


/****************************************************************
 * Called by check_break(state, breaks) in main.c.  This function
 * updates the number of occurances remaining for the breakpoint
 * at address PC.  If it has reached its last occurance, it is
 * removed from the list of active breakpoints.
 ****************************************************************/
void decrementTempBrkpt(int PC)
{
    int   i, numItems=0, address=0, occurances=0;
    char  addr[40], text[80];

    /* Find the breakpoint corresponding to PC */
    numItems = (int)xv_get(Sim6811_winBreakpoints->listBreakPts, 
			   PANEL_LIST_NROWS, NULL);

    for (i=0; i<numItems; i++) {
	/* Compare address to PC */
	strcpy(addr,(char*)xv_get(Sim6811_winBreakpoints->listBreakPts, 
				   PANEL_LIST_STRING, i, NULL));

	sscanf(addr, "0x%x (Temporary: %d", &address, &occurances);

	if (address == PC) break;
    }

    /* If no matching breakpoint was found, simply exit. */
    if (i == numItems) {
	return;
    }
    else {

	/* Otherwise, decrement the number of remaining occurances */
	--occurances;

	if (occurances > 0) {
	    /* Change the breakpoint's list entry to reflect
	       the number of remaining occurances. */
	    sprintf(text, "0x%04x (Temporary: %d times)", 
		    address, occurances);

	    xv_set(Sim6811_winBreakpoints->listBreakPts, 
		   PANEL_LIST_DELETE, i, 
		   PANEL_LIST_INSERT, i, 
		   PANEL_LIST_STRING, i, text, 
		   PANEL_LIST_SORT, PANEL_FORWARD, NULL);
	}
	else {
	    /* Remove the breakpoint from the list */
	    xv_set(Sim6811_winBreakpoints->listBreakPts, 
		   PANEL_LIST_DELETE, i, 
		   PANEL_LIST_SORT, PANEL_FORWARD, NULL);

	    if (numItems-1 == 0) {
		/* Deactivate the "Update" and "Delete" buttons */
		xv_set(Sim6811_winBreakpoints->btnBreakPtsUpdate, 
		       PANEL_INACTIVE, TRUE, NULL); 
		xv_set(Sim6811_winBreakpoints->btnBreakPtsDelete, 
		       PANEL_INACTIVE, TRUE, NULL); 
	    }
	}
    }
}
